/*    This program solves a minesweeper game.

VVVVVVVVVV   please fill in these fields VVVVVVVVVVVVV
24 September 2012	               <Sahil Gupta>   */

#include "minefield.h"
#define OFFSET(r, c, numcols)  ((r)*(numcols)+(c))

/*
This function checks whether the sqaure is on the corner or not.
*/

void cornerChecker(int i, int j, int fs, int *i_adjacent, int *j_adjacent, int *pAdjacentCount) {
    if (i == 0 && j == 0) {
        i_adjacent[0] = 0;
        i_adjacent[1] = 1;
        i_adjacent[2] = 1;

        j_adjacent[0] = 1;
        j_adjacent[1] = 0;
        j_adjacent[2] = 1;

        *pAdjacentCount = 3;
    } else if (i == (fs - 1) && j == 0) {
        i_adjacent[0] = fs - 1;
        i_adjacent[1] = fs - 2;
        i_adjacent[2] = fs - 2;

        j_adjacent[0] = 1;
        j_adjacent[1] = 0;
        j_adjacent[2] = 1;

        *pAdjacentCount = 3;
    } else if (i == (fs - 1) && j == (fs - 1)) {
        i_adjacent[0] = fs - 1;
        i_adjacent[1] = fs - 2;
        i_adjacent[2] = fs - 2;

        j_adjacent[0] = fs - 2;
        j_adjacent[1] = fs - 1;
        j_adjacent[2] = fs - 2;

        *pAdjacentCount = 3;
    } else if (i == 0 && j == (fs - 1)) {
        i_adjacent[0] = 0;
        i_adjacent[1] = 1;
        i_adjacent[2] = 1;

        j_adjacent[0] = fs - 2;
        j_adjacent[1] = fs - 1;
        j_adjacent[2] = fs - 2;

        *pAdjacentCount = 3;
    }
}

/*
This function checks whether the sqaure is on the boundary/inside or not.
*/

void boundaryOrInsideChecker(int i, int j, int fs, int *i_adjacent, int *j_adjacent, int *pAdjacentCount) {

    if (j == 0) {
        i_adjacent[0] = i;
        j_adjacent[0] = 1;
        i_adjacent[1] = i-1;
        j_adjacent[1] = 0;
        i_adjacent[2] = i+1;
        j_adjacent[2] = 0;
        i_adjacent[3] = i-1;
        j_adjacent[3] = 1;
        i_adjacent[4] = i+1;
        j_adjacent[4] = 1;

        *pAdjacentCount = 5;
    } else if (j == (fs - 1)) {
        i_adjacent[0] = i;
        i_adjacent[1] = i-1;
        i_adjacent[2] = i+1;
        i_adjacent[3] = i-1;
        i_adjacent[4] = i+1;

        j_adjacent[0] = fs-2;
        j_adjacent[1] = fs-1;
        j_adjacent[2] = fs-1;
        j_adjacent[3] = fs-2;
        j_adjacent[4] = fs-2;

        *pAdjacentCount = 5;
    } else if (i == 0) {
        i_adjacent[0] = 0;
        i_adjacent[1] = 0;
        i_adjacent[2] = 1;
        i_adjacent[3] = 1;
        i_adjacent[4] = 1;

        j_adjacent[0] = j-1;
        j_adjacent[1] = j+1;
        j_adjacent[2] = j;
        j_adjacent[3] = j-1;
        j_adjacent[4] = j+1;

        *pAdjacentCount = 5;
    } else if (i == (fs - 1)) {
        i_adjacent[0] = fs-1;
        i_adjacent[1] = fs-1;
        i_adjacent[2] = fs-2;
        i_adjacent[3] = fs-2;
        i_adjacent[4] = fs-2;

        j_adjacent[0] = j-1;
        j_adjacent[1] = j+1;
        j_adjacent[2] = j;
        j_adjacent[3] = j-1;
        j_adjacent[4] = j+1;

        *pAdjacentCount = 5;
    } else {
        i_adjacent[0] = i-1;
        i_adjacent[1] = i+1;
        i_adjacent[2] = i;
        i_adjacent[3] = i;
        i_adjacent[4] = i-1;
        i_adjacent[5] = i-1;
        i_adjacent[6] = i+1;
        i_adjacent[7] = i+1;

        j_adjacent[0] = j;
        j_adjacent[1] = j;
        j_adjacent[2] = j-1;
        j_adjacent[3] = j+1;
        j_adjacent[4] = j-1;
        j_adjacent[5] = j+1;
        j_adjacent[6] = j-1;
        j_adjacent[7] = j+1;

        *pAdjacentCount = 8;
    }
}

/*
This method checks the two rules mentioned to solve the minesweeper game.
*/

void mineDetector(int i, int j, int *mineArray, int fs, int *ptrChangeTracker1) {

    int val = mineArray[OFFSET(i, j, fs)];
    int rval, cval, tempVal;
    int n;

    int i_adjacent[8];
    int j_adjacent[8];

    int adjacentCount, flagCount = 0, unOpenedCounter = 0;
    int *pAdjacentCount = &adjacentCount;

    cornerChecker(i, j, fs, i_adjacent, j_adjacent, pAdjacentCount);
    boundaryOrInsideChecker(i, j, fs, i_adjacent, j_adjacent, pAdjacentCount);

    n = 0;
    while (n < adjacentCount) {
        tempVal = mineArray[OFFSET(i_adjacent[n], j_adjacent[n], fs)];

        if (tempVal == -1) {
            flagCount++;
        }
        if (tempVal == -2) {//-2 is for un-opened squares since intialized in array.
            unOpenedCounter++;
        }
        n++;
    }

    n = 0;
    if ((val == (unOpenedCounter + flagCount))) {
        while (n < adjacentCount) {
            rval = i_adjacent[n];
            cval = j_adjacent[n];
            tempVal = mineArray[OFFSET(rval, cval, fs)];
            if (tempVal == -2) {

                flag(rval, cval);
                mineArray[OFFSET(rval, cval, fs)] = -1;
                (*ptrChangeTracker1)++;
            }
            n++;
        }
    }

    n = 0;
    if (val == 0 || val == flagCount) {
        while (n < adjacentCount) {
            rval = i_adjacent[n];
            cval = j_adjacent[n];
            tempVal = mineArray[OFFSET(rval, cval, fs)];

            if (tempVal == -2) {

                mineArray[OFFSET(rval, cval, fs)] = open(rval, cval);

                (*ptrChangeTracker1)++;
            }
            n++;
        }
    }
}

/*
This method contains the main algorithm to solve the minesweeper game.
*/

void solver(int fs, int mn) {

    int fieldSize = fs;
    int mineFieldSize = fieldSize*fieldSize;

    int i = 0, j = 0;
    int mineArray[mineFieldSize];

    while (i < mineFieldSize) {
        mineArray[i] = -2;
        i++;
    }

    int changeTracker1 = 0, changeTracker2 = 0;
    /*	These will help us in checking whether a change is made to*/
    /*	 the state of the Mine Array when updating it.*/

    int *ptrChangeTracker1 = &changeTracker1;

    int val = guess(0, 0);
    mineArray[OFFSET(0, 0, fs)] = val;

    while (1) {
        int Bool = 0;
        changeTracker2 = changeTracker1;

        i = 0;
        while (i < fs) {
            j = 0;
            while (j < fs) {
                val = mineArray[OFFSET(i, j, fs)];
                if (val >= 0) {
                    mineDetector(i, j, mineArray, fs, ptrChangeTracker1);
                }
                j++;
            }
            i++;
        }

        if (changeTracker1 == changeTracker2) {
            i = 0;
            while ((i < fs) && !Bool) {
                j = 0;
                while (j < fs) {
                    val = mineArray[OFFSET(i, j, fs)];
                    if (val == -2) {
                        mineArray[OFFSET(i, j, fs)] = guess(i, j);
                        changeTracker1++;
                        Bool = 1;
                        break;
                    }
                    j++;
                }
                i++;
            }

        }

    }
    display_field_discovered_so_far();
}
